@@ -12,3 +12,5 @@ RELATIVE_VARS = ['pr']
 ABSOLUTE_VARS = ['tasmax', 'tasmin']
 APHYSICAL_TEMP_HIGH = 353  # Kelvin
 APHYSICAL_TEMP_LOW = 0
+APHYSICAL_PRECIP_HIGH = 3000
+APHYSICAL_PRECIP_LOW = 0
