@@ -6,10 +6,8 @@ import json
 import os
 import warnings
 from dataclasses import asdict
-from datetime import timedelta
 from pathlib import PosixPath
 
-import datatree
 import datatree as dt
 import fsspec
 import pandas as pd
@@ -17,25 +15,30 @@ import rechunker
 import xarray as xr
 import zarr
 from carbonplan_data.metadata import get_cf_global_attrs
-from carbonplan_data.utils import set_zarr_encoding
+from carbonplan_data.utils import set_zarr_encoding as set_web_zarr_encoding
 from ndpyramid import pyramid_regrid
 from prefect import task
+from prefect.triggers import any_failed
 from upath import UPath
 from xarray_schema import DataArraySchema, DatasetSchema
 from xarray_schema.base import SchemaError
 
-from cmip6_downscaling import __version__ as version, config
-from cmip6_downscaling.data.cmip import get_gcm
-from cmip6_downscaling.data.observations import open_era5
-from cmip6_downscaling.methods.common.containers import RunParameters
-from cmip6_downscaling.methods.common.utils import (
+from ... import __version__ as version, config
+from ...data.cmip import get_gcm
+from ...data.observations import open_era5
+from ...utils import str_to_hash
+from .containers import RunParameters, TimePeriod
+from .utils import (
+    blocking_to_zarr,
     calc_auspicious_chunks_dict,
+    is_cached,
     resample_wrapper,
+    set_zarr_encoding,
     subset_dataset,
-    zmetadata_exists,
+    validate_zarr_store,
 )
-from cmip6_downscaling.utils import str_to_hash
 
+xr.set_options(keep_attrs=True)
 warnings.filterwarnings(
     "ignore",
     "(.*) filesystem path not explicitly implemented. falling back to default implementation. This filesystem may not be tested",
@@ -49,13 +52,13 @@ results_dir = UPath(config.get("storage.results.uri")) / version
 use_cache = config.get('run_options.use_cache')
 
 
-@task(log_stdout=True, max_retries=3, retry_delay=timedelta(seconds=5))
+@task(log_stdout=True)
 def make_run_parameters(**kwargs) -> RunParameters:
     """Prefect task wrapper for RunParameters"""
     return RunParameters(**kwargs)
 
 
-@task(log_stdout=True, max_retries=3, retry_delay=timedelta(seconds=5))
+@task(log_stdout=True)
 def get_obs(run_parameters: RunParameters) -> UPath:
     """Task to return observation data subset from input parameters.
 
@@ -69,41 +72,38 @@ def get_obs(run_parameters: RunParameters) -> UPath:
     UPath
         Path to subset observation dataset.
     """
-
-    title = "obs ds: {obs}_{variable}_{latmin}_{latmax}_{lonmin}_{lonmax}_{train_dates[0]}_{train_dates[1]}".format(
-        **asdict(run_parameters)
-    )
-    ds_hash = str_to_hash(
-        "{obs}_{variable}_{latmin}_{latmax}_{lonmin}_{lonmax}_{train_dates[0]}_{train_dates[1]}".format(
-            **asdict(run_parameters)
-        )
+    feature_string = '_'.join(run_parameters.features)
+    frmt_str = "{obs}_{feature_string}_{latmin}_{latmax}_{lonmin}_{lonmax}_{train_dates[0]}_{train_dates[1]}".format(
+        **asdict(run_parameters), feature_string=feature_string
     )
+    title = f"obs ds: {frmt_str}"
+    ds_hash = str_to_hash(frmt_str)
     target = intermediate_dir / 'get_obs' / ds_hash
 
-    if use_cache and zmetadata_exists(target):
+    if use_cache and is_cached(target):
         print(f'found existing target: {target}')
         return target
 
-    ds = open_era5(run_parameters.variable, run_parameters.train_period)
+    ds = open_era5(run_parameters.features, run_parameters.train_period)
 
     subset = subset_dataset(
         ds,
-        run_parameters.variable,
+        run_parameters.features,
         run_parameters.train_period.time_slice,
         run_parameters.bbox,
         chunking_schema={'time': 365, 'lat': 150, 'lon': 150},
     )
-
-    if run_parameters.variable != 'pr':
-        del subset[run_parameters.variable].encoding['chunks']
+    subset = ds.chunk({'time': 365, 'lat': 150, 'lon': 150})
+    for key in subset.variables:
+        subset[key].encoding = {}
 
     subset.attrs.update({'title': title}, **get_cf_global_attrs(version=version))
-    store = subset.to_zarr(target, mode='w', compute=False)
-    store.compute(retries=5)
+    subset = set_zarr_encoding(subset)
+    blocking_to_zarr(ds=subset, target=target, validate=True, write_empty_chunks=True)
     return target
 
 
-@task(log_stdout=True, max_retries=3, retry_delay=timedelta(seconds=5))
+@task(log_stdout=True)
 def get_experiment(run_parameters: RunParameters, time_subset: str) -> UPath:
     """Prefect task that returns cmip GCM data from input run parameters.
 
@@ -112,56 +112,88 @@ def get_experiment(run_parameters: RunParameters, time_subset: str) -> UPath:
     run_parameters : RunParameters
         RunParameter dataclass defined in common/conatiners.py. Constructed from prefect parameters.
     time_subset : str
-        String describing time subset request. Either 'train_period' or 'predict_period'
+        String describing time subset request. Either 'train_period', 'predict_period', or 'both'.
 
     Returns
     -------
     UPath
         UPath to experiment dataset.
     """
-    time_period = getattr(run_parameters, time_subset)
-    frmt_str = "{model}_{member}_{scenario}_{variable}_{latmin}_{latmax}_{lonmin}_{lonmax}_{time_period.start}_{time_period.stop}".format(
-        time_period=time_period, **asdict(run_parameters)
-    )
+
+    if time_subset == 'both':
+        time_period = TimePeriod(
+            start=str(
+                min(
+                    int(run_parameters.train_period.start), int(run_parameters.predict_period.start)
+                )
+            ),
+            stop=str(
+                max(int(run_parameters.train_period.stop), int(run_parameters.predict_period.stop))
+            ),
+        )
+    else:
+        time_period = getattr(run_parameters, time_subset)
+
+    features = getattr(run_parameters, 'features')
+    if features:
+        feature_string = '_'.join(features)
+        frmt_str = "{model}_{member}_{scenario}_{feature_string}_{latmin}_{latmax}_{lonmin}_{lonmax}_{time_period.start}_{time_period.stop}".format(
+            time_period=time_period, **asdict(run_parameters), feature_string=feature_string
+        )
+
+    else:
+        frmt_str = "{model}_{member}_{scenario}_{variable}_{latmin}_{latmax}_{lonmin}_{lonmax}_{time_period.start}_{time_period.stop}".format(
+            time_period=time_period, **asdict(run_parameters)
+        )
+
+    if int(time_period.start) < 2015 and run_parameters.scenario != 'historical':
+        scenarios = ['historical', run_parameters.scenario]
+    else:
+        scenarios = [run_parameters.scenario]
 
     title = f"experiment ds: {frmt_str}"
     ds_hash = str_to_hash(frmt_str)
     target = intermediate_dir / 'get_experiment' / ds_hash
 
-    print(target)
-    if use_cache and zmetadata_exists(target):
+    if use_cache and is_cached(target):
         print(f'found existing target: {target}')
         return target
 
-    ds = get_gcm(
-        scenario=run_parameters.scenario,
-        member_id=run_parameters.member,
-        table_id=run_parameters.table_id,
-        grid_label=run_parameters.grid_label,
-        source_id=run_parameters.model,
-        variable=run_parameters.variable,
-        time_slice=time_period.time_slice,
-    )
+    mode = 'w'
+    for feature in run_parameters.features:
+        ds_list = []
+        for s in scenarios:
+            ds_list.append(
+                get_gcm(
+                    scenario=s,
+                    member_id=run_parameters.member,
+                    table_id=run_parameters.table_id,
+                    grid_label=run_parameters.grid_label,
+                    source_id=run_parameters.model,
+                    variable=feature,
+                    time_slice=time_period.time_slice,
+                )
+            )
+        ds = xr.concat(ds_list, dim='time')
+        subset = subset_dataset(ds, feature, time_period.time_slice, run_parameters.bbox)
+        # Note: dataset is chunked into time:365 chunks to standardize leap-year chunking.
+        subset = subset.chunk({'time': 365})
+        for key in subset.variables:
+            subset[key].encoding = {}
+        subset.attrs.update({'title': title}, **get_cf_global_attrs(version=version))
+        subset = set_zarr_encoding(subset)
+        subset[[feature]].to_zarr(target, mode=mode)
+        mode = 'a'
 
-    subset = subset_dataset(
-        ds, run_parameters.variable, time_period.time_slice, run_parameters.bbox
-    )
-
-    # Note: dataset is chunked into time:365 chunks to standardize leap-year chunking.
-    subset = subset.chunk({'time': 365})
-    if run_parameters.variable != 'pr':
-        del subset[run_parameters.variable].encoding['chunks']
-    subset.attrs.update({'title': title}, **get_cf_global_attrs(version=version))
-    subset.to_zarr(target, mode='w')
     return target
 
 
-@task(log_stdout=True, max_retries=3, retry_delay=timedelta(seconds=5))
+@task(log_stdout=True)
 def rechunk(
     path: UPath,
     pattern: str = None,
     template: UPath = None,
-    max_mem: str = "2GB",
+    max_mem: str = "5GB",
 ) -> UPath:
     """Use `rechunker` package to adjust chunks of dataset to a form
     conducive for your processing.
@@ -178,7 +210,7 @@ def rechunk(
         The path to the file you want to use as a chunking template. The utility will grab the chunk sizes and use them as the chunk
         target to feed to rechunker.
     max_mem : str
-        The memory available for rechunking steps. Must look like "2GB". Optional, default is 2GB.
+        The memory available for rechunking steps. Must look like "2GB". Optional, default is 5GB.
 
     Returns
     -------
@@ -196,12 +228,11 @@ def rechunk(
     task_hash = str_to_hash(str(path) + pattern_string + str(template) + max_mem)
     target = intermediate_dir / 'rechunk' / task_hash
     path_tmp = scratch_dir / 'rechunk' / task_hash
-    print(f'writing rechunked dataset to {target}')
-    print(target)
+
     target_store = fsspec.get_mapper(str(target))
     temp_store = fsspec.get_mapper(str(path_tmp))
 
-    if use_cache and zmetadata_exists(target):
+    if use_cache and is_cached(target):
         print(f'found existing target: {target}')
         # if we wanted to check that it was chunked correctly we could put this down below where
         # the target_schema is validated. but that requires us going through the development
@@ -233,7 +264,6 @@ def rechunk(
             chunk_dims = config.get(f"chunk_dims.{pattern}")
             for dim in chunk_def:
                 if dim not in chunk_dims:
-                    print('correcting dim')
                     # override the chunksize of those unchunked dimensions to be the complete length (like passing chunksize=-1
                     chunk_def[dim] = len(ds[dim])
     elif pattern is not None:
@@ -270,17 +300,24 @@ def rechunk(
         max_mem=max_mem,
         target_store=target_store,
         temp_store=temp_store,
+        target_options={
+            k: {'compressor': zarr.Blosc(clevel=1), 'write_empty_chunks': True} for k in chunks_dict
+        },
+        temp_options={k: {'compressor': None, 'write_empty_chunks': True} for k in chunks_dict},
+        executor='dask',
     )
 
     rechunk_plan.execute()
 
     # consolidate_metadata here since when it comes out of rechunker it isn't consolidated.
     zarr.consolidate_metadata(target_store)
+    validate_zarr_store(target_store)
+
     temp_store.clear()
     return target
 
 
-@task(log_stdout=True, max_retries=3, retry_delay=timedelta(seconds=5))
+@task(log_stdout=True)
 def time_summary(ds_path: UPath, freq: str) -> UPath:
     """Prefect task to create resampled data. Takes mean of `tasmax` and `tasmin` and sum of `pr`.
 
@@ -298,9 +335,9 @@ def time_summary(ds_path: UPath, freq: str) -> UPath:
     """
 
     ds_hash = str_to_hash(str(ds_path) + freq)
-    target = intermediate_dir / 'time_summary' / ds_hash
-    print(target)
-    if use_cache and zmetadata_exists(target):
+    target = results_dir / 'time_summary' / ds_hash
+
+    if use_cache and is_cached(target):
         print(f'found existing target: {target}')
         return target
 
@@ -309,12 +346,13 @@ def time_summary(ds_path: UPath, freq: str) -> UPath:
     out_ds = resample_wrapper(ds, freq=freq)
 
     out_ds.attrs.update({'title': 'time_summary'}, **get_cf_global_attrs(version=version))
-    out_ds.to_zarr(target, mode='w')
+    out_ds = set_zarr_encoding(out_ds)
+    blocking_to_zarr(ds=out_ds, target=target, validate=True, write_empty_chunks=True)
 
     return target
 
 
-@task(log_stdout=True, max_retries=3, retry_delay=timedelta(seconds=5))
+@task(log_stdout=True)
 def get_weights(*, run_parameters, direction, regrid_method="bilinear"):
     """Retrieve pre-generated regridding weights.
 
@@ -344,11 +382,10 @@ def get_weights(*, run_parameters, direction, regrid_method="bilinear"):
         .iloc[0]
         .path
     )
-    print(path)
     return path
 
 
-@task(log_stdout=True, max_retries=3, retry_delay=timedelta(seconds=5))
+@task(log_stdout=True)
 def get_pyramid_weights(*, run_parameters, levels: int, regrid_method: str = "bilinear"):
     """Retrieve pre-generated regridding pyramids weights.
 
@@ -367,16 +404,19 @@ def get_pyramid_weights(*, run_parameters, levels: int, regrid_method: str = "bi
         Path to pyramid weights file.
     """
     weights = pd.read_csv(config.get('weights.downscaled_pyramid_weights.uri'))
-    print(weights)
     path = (
         weights[(weights.regrid_method == regrid_method) & (weights.levels == levels)].iloc[0].path
     )
-    print(path)
     return path
 
 
-@task(log_stdout=True, max_retries=3, retry_delay=timedelta(seconds=5))
-def regrid(source_path: UPath, target_grid_path: UPath, weights_path: UPath = None) -> UPath:
+@task(log_stdout=True)
+def regrid(
+    source_path: UPath,
+    target_grid_path: UPath,
+    weights_path: UPath = None,
+    pre_chunk_def: dict = None,
+) -> UPath:
     """Task to regrid a dataset to target grid.
 
     Parameters
@@ -398,18 +438,21 @@ def regrid(source_path: UPath, target_grid_path: UPath, weights_path: UPath = No
 
     ds_hash = str_to_hash(str(source_path) + str(target_grid_path))
     target = intermediate_dir / 'regrid' / ds_hash
-    print(target)
 
-    if use_cache and zmetadata_exists(target):
+    if use_cache and is_cached(target):
         print(f'found existing target: {target}')
         return target
+
     source_ds = xr.open_zarr(source_path)
     target_grid_ds = xr.open_zarr(target_grid_path)
+
+    if pre_chunk_def is not None:
+        source_ds = source_ds.chunk(**pre_chunk_def)
+
     if weights_path:
         from ndpyramid.regrid import _reconstruct_xesmf_weights
 
         weights = _reconstruct_xesmf_weights(xr.open_zarr(weights_path))
-        print(weights_path)
         regridder = xe.Regridder(
             source_ds,
             target_grid_ds,
@@ -432,7 +475,9 @@ def regrid(source_path: UPath, target_grid_path: UPath, weights_path: UPath = No
     regridded_ds.attrs.update(
         {'title': source_ds.attrs['title']}, **get_cf_global_attrs(version=version)
     )
-    regridded_ds.to_zarr(target, mode='w')
+
+    regridded_ds = set_zarr_encoding(regridded_ds)
+    blocking_to_zarr(ds=regridded_ds, target=target, validate=True, write_empty_chunks=True)
     return target
 
 
@@ -476,7 +521,7 @@ def _pyramid_postprocess(dt: dt.DataTree, levels: int, other_chunks: dict = None
             dt[slevel].ds['date_str'] = dt[slevel].ds['date_str'].chunk(-1)
 
         # set dataset encoding
-        dt[slevel].ds = set_zarr_encoding(
+        dt[slevel].ds = set_web_zarr_encoding(
             dt[slevel].ds, codec_config={"id": "zlib", "level": 1}, float_dtype="float32"
         )
         for var in ['time', 'time_bnds']:
@@ -488,9 +533,9 @@ def _pyramid_postprocess(dt: dt.DataTree, levels: int, other_chunks: dict = None
     return dt
 
 
-@task(log_stdout=True, max_retries=3, retry_delay=timedelta(seconds=5))
+@task(log_stdout=True)
 def pyramid(
-    ds_path: UPath, weights_pyramid_path: str, levels: int = 2, other_chunks: dict = None
+    ds_path: UPath, weights_pyramid_path: str = None, levels: int = 2, other_chunks: dict = None
 ) -> UPath:
     '''Task to create a data pyramid from an xarray Dataset
 
@@ -515,7 +560,7 @@ def pyramid(
     ds_hash = str_to_hash(str(ds_path) + str(levels) + str(other_chunks))
     target = results_dir / 'pyramid' / ds_hash
 
-    if use_cache and zmetadata_exists(target):
+    if use_cache and is_cached(target):
         print(f'found existing target: {target}')
         return target
 
@@ -525,7 +570,10 @@ def pyramid(
 
     ds.attrs.update({'title': ds.attrs['title']}, **get_cf_global_attrs(version=version))
     target_pyramid = dt.open_datatree('az://static/target-pyramid', engine='zarr')
-    weights_pyramid = datatree.open_datatree(weights_pyramid_path, engine='zarr')
+    if weights_pyramid_path is not None:
+        weights_pyramid = dt.open_datatree(weights_pyramid_path, engine='zarr')
+    else:
+        weights_pyramid = None
     # create pyramid
     dta = pyramid_regrid(
         ds,
@@ -535,12 +583,19 @@ def pyramid(
         regridder_kws={'ignore_degenerate': True},
     )
 
+    dta = _pyramid_postprocess(dta, levels=levels, other_chunks=other_chunks)
+
     # write to target
+    for child in dta.children.values():
+        for variable in child.ds.data_vars:
+            child[variable].encoding['write_empty_chunks'] = True
+
     dta.to_zarr(target, mode='w')
+    validate_zarr_store(target)
     return target
 
 
-@task(log_stdout=True, max_retries=3, retry_delay=timedelta(seconds=5))
+@task(log_stdout=True)
 def run_analyses(ds_path: UPath, run_parameters: RunParameters) -> UPath:
     """Prefect task to run the analyses on results from a downscaling run.
 
@@ -605,24 +660,14 @@ def run_analyses(ds_path: UPath, run_parameters: RunParameters) -> UPath:
     return executed_notebook_path
 
 
-@task(log_stdout=True, max_retries=3, retry_delay=timedelta(seconds=5))
-def finalize(path_dict: dict, run_parameters: RunParameters):
-    """Prefect task to finalize the downscaling run.
-
-    Parameters
-    ----------
-    path_dict : dict
-        Dictionary of paths to write to
-    run_parameters : RunParameters
-        Downscaling run parameter container
-
-    """
+def _finalize(run_parameters, kind='runs', **paths):
+    path_dict = dict(**paths)
 
     now = datetime.datetime.utcnow().isoformat()
-    target1 = results_dir / 'runs' / run_parameters.run_id / f'{now}.json'
-    target2 = results_dir / 'runs' / run_parameters.run_id / 'latest.json'
-    print(target1)
-    print(target2)
+    target1 = results_dir / kind / run_parameters.run_id / f'{now}.json'
+    target2 = results_dir / kind / run_parameters.run_id / 'latest.json'
+    print(f'finalize 1: {target1}')
+    print(f'finalize 2: {target2}')
 
     out = {'parameters': asdict(run_parameters)}
     out['attrs'] = get_cf_global_attrs(version=version)
@@ -633,3 +678,32 @@ def finalize(path_dict: dict, run_parameters: RunParameters):
 
     with target2.open(mode='w') as f:
         json.dump(out, f, indent=2)
+
+
+@task(log_stdout=True)
+def finalize(run_parameters: RunParameters = None, **paths):
+    """Prefect task to finalize the downscaling run.
+
+    Parameters
+    ----------
+    run_parameters : RunParameters
+        Downscaling run parameter container
+    paths : dict
+        Dictionary of paths to write result file
+    """
+    _finalize(run_parameters, kind='runs', **paths)
+
+
+@task(log_stdout=True, trigger=any_failed)
+def finalize_on_failure(run_parameters: RunParameters = None, **paths):
+    """Prefect task to finalize a failed downscaling run.
+
+    Parameters
+    ----------
+    run_parameters : RunParameters
+        Downscaling run parameter container
+    paths : dict
+        Dictionary of paths to write to result file
+    """
+
+    _finalize(run_parameters, kind='failed-runs', **paths)
